from collections import OrderedDict
import logging

from graphql import (
    GraphQLSchema,
    GraphQLObjectType,
    GraphQLField,
    GraphQLString,
    GraphQLEnumType,
    GraphQLInt,
    GraphQLFloat,
    GraphQLBoolean,
    GraphQLID,
    GraphQLNonNull,
    GraphQLList,
    GraphQLArgument,
)

from graphql.type.definition import (
    GraphQLEnumValue,
    GraphQLInterfaceType,
)

GRAPHQL_SCALARS = {
    'Int': GraphQLInt,
    'Float': GraphQLFloat,
    'String': GraphQLString,
    'Boolean': GraphQLBoolean,
    'ID': GraphQLID
}


LOGGER = logging.getLogger("idlewild")


class UnregisteredError(Exception):
    pass


class SchemaMappingsInvalid(Exception):
    pass


class Builder:

    def __init__(
            self, resolver_map=None,
            interface_resolver_map=None):
        self.types = {}
        self.enums = {}
        self.interfaces = {}
        self.schema = None
        self.resolver_map = (
            resolver_map
            if resolver_map is not None else {}
        )
        self.interface_resolver_map = (
            interface_resolver_map
            if interface_resolver_map is not None else {}
        )

    def build(self, items):
        root = None  # root has to be last
        for item in items:
            item_id, *_ = item
            if item_id == 'SCHEMADEF':
                root = item
            else:
                self.eval(item)
        self.eval(root)

    def eval(self, item):
        item_id, *_ = item
        if item_id == 'TYPE':
            return self.register_type(item)
        if item_id == 'INTERFACE':
            return self.register_interface(item)
        if item_id == 'ENUM':
            return self.register_enum(item)
        if item_id == 'SCHEMADEF':
            return self.register_schemadef(item)
        raise ValueError(
            'item {} did was not correctly defined'.format(item)
        )

    def register_enum(self, item):
        _, enum_name, enum_values = item

        def create_values(enum_values):
            value_dict = OrderedDict()
            for count, name in list(enumerate(enum_values)):
                value_dict[name] = GraphQLEnumValue(count)
            return value_dict

        enum = GraphQLEnumType(
            name=enum_name,
            values=create_values(enum_values)
        )

        self.enums[enum_name] = enum
        LOGGER.info('Registered ENUM {}'.format(enum_name))

    def _resolve_base_type(self, type_ref):
        base_type = {
            **GRAPHQL_SCALARS,
            **self.types,
            **self.interfaces,
            **self.enums,
            **self.interfaces}.get(type_ref)
        if base_type is None:
            LOGGER.error('Could not resolve {}'.format(type_ref))
            raise UnregisteredError
        return base_type

    def _extract_type_info(self, field_type_info):
        is_list, type_info = field_type_info
        _, is_nullable, type_ref = type_info
        return is_list, is_nullable, type_ref

    def _build_field_type(self, field_type_info):
        is_list, is_nullable, type_ref = (
            self._extract_type_info(field_type_info)
        )
        base_type = self._resolve_base_type(type_ref)

        if is_nullable == 'NON-NULLABLE':
            base_type = GraphQLNonNull(base_type)
        if is_list == 'GRAPHQL_LIST':
            base_type = GraphQLList(base_type)
        if is_list == 'GRAPHQL_NON-NULLABLE_LIST':
            base_type = GraphQLNonNull(GraphQLList(base_type))

        return base_type

    def _build_args(self, args_list):
        target_args = []
        for arg in args_list:
            name_and_args, field_type_info = arg
            _, name, *_ = name_and_args
            target_args.append(
                (name, GraphQLArgument(
                    self._build_field_type(field_type_info)
                 )))
        return dict(target_args)

    def _build_fields(self, fields):
        target_fields = []
        for field in fields:
            name_and_args, field_type_info = field
            _, name, _, args_list = name_and_args
            target_fields.append(
                (name, GraphQLField(
                    self._build_field_type(field_type_info),
                    args=self._build_args(args_list),
                    resolver=self.resolver_map.get(name),
                 )))
        return dict(target_fields)

    def register_interface(self, item):
        _, interface_name, interface_fields = item

        interface = GraphQLInterfaceType(
            name=interface_name,
            fields=lambda: dict(self._build_fields(interface_fields)),
            resolve_type=lambda x, y: self._resolve_base_type(
                self.interface_resolver_map.get(interface_name)(x, y)
            )
        )

        self.interfaces[interface_name] = interface
        LOGGER.info('Registered INTERFACE {}'.format(interface_name))

    def register_type(self, item):
        _, type_name, implements, field_list = item

        target_type = GraphQLObjectType(
            name=type_name,
            fields=lambda: self._build_fields(field_list),
            interfaces=lambda: (
                (self._resolve_base_type(implements),)
                if implements else None
             )
        )
        self.types[type_name] = target_type
        LOGGER.info('Registered TYPE {}'.format(type_name))

    def register_schemadef(self, item):
        _, _, definition_list = item  # TODO: tidy this
        for definition in definition_list:
            root_id_info, root_type_info = definition
            _, root_item_info = root_type_info
            _, root_name = root_type_info
            _, _, root_target = root_name
            self.schema = GraphQLSchema(
                query=self._resolve_base_type(root_target)
            )
        LOGGER.info('Created root schema')



import time
import logging
import sys

import idlewild.parser as parser
from idlewild.builder import Builder


LOGGER = logging.getLogger("idlewild")


def setup_logging(logfile_name=None):
    logger = logging.getLogger()
    ch = (
        logging.FileHandler(logfile_name) if logfile_name
        else logging.StreamHandler()
    )
    logger.setLevel(logging.INFO)
    ch.setLevel(logging.INFO)
    format_string = '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
    formatter = logging.Formatter(format_string)
    ch.setFormatter(formatter)
    logger.addHandler(ch)


def parse_idl(idl):
    try:
        LOGGER.info('Starting parsing')
        start = time.time()
        nodes = parser.parse(idl)
        LOGGER.info('Parsing complete in [{}] ms'.format(
            (time.time()-start) / 3600
        ))
        return nodes
    except SyntaxError as se:
        LOGGER.error('Parse Error: {0}'.format(se))


def build_idl_schema(
        nodes,
        resolver_mappings=None,
        interface_resolver_mappings=None):
    builder = Builder(
        resolver_mappings, interface_resolver_mappings
    )
    LOGGER.info('Starting build')
    start_build = time.time()
    builder.build(nodes)
    LOGGER.info('Schema built in [{}] ms'.format(
        (time.time()-start_build) / 3600
    ))
    return builder.schema


def parse_and_build_idl(
        idl,
        resolver_mappings=None,
        interface_resolver_mappings=None):
    nodes = parse_idl(idl)
    return build_idl_schema(
        nodes, resolver_mappings, interface_resolver_mappings
    )


def parse_idl_file(filename):
    LOGGER.info('Loading IDL from {}'.format(filename))
    with open(filename) as f:
        data = f.read()
        return parse_idl(data)


def parse_and_build_idl_file(
        filename,
        resolver_mappings=None,
        interface_resolver_mappings=None):
    LOGGER.info('Building schema from {}'.format(filename))
    nodes = parse_idl_file(filename)
    return build_idl_schema(
        nodes, resolver_mappings, interface_resolver_mappings
    )


if __name__ == '__main__':
    if len(sys.argv) == 2:
        setup_logging()
        schema = parse_and_build_idl_file(sys.argv[1])
        LOGGER.info('Parsed and validated schema')
    else:
        print('Usage: schematools [idl-file]')
        raise SystemExit



import ply.lex as lex

DEBUG_STATUS = False  # show the low level LALR debug
PARSER_OUT = False  # produce the parser.out for debug
DEBUG_TABLES = False  # show the parsing tables

OPTIMIZE = False

keywords = (
    'TYPE', 'ENUM', 'IMPLEMENTS', 'INTERFACE', 'SCHEMA',
)

tokens = keywords + (
    'ID', 'COLON', 'BANG', 'LSQUARE', 'RSQUARE',
    'LBRACE', 'RBRACE', 'COMMA', 'LPAREN', 'RPAREN',
)

t_ignore = ' \t\n'

t_COMMA = r'\,'
t_COLON = r'\:'
t_BANG = r'\!'
t_LSQUARE = r'\['
t_RSQUARE = r'\]'
t_LBRACE = r'\{'
t_RBRACE = r'\}'
t_LPAREN = r'\('
t_RPAREN = r'\)'


def t_error(t):
    raise SyntaxError("Unknown symbol %r" % (t.value[0],))
    print("Skipping", repr(t.value[0]))
    t.lexer.skip(1)


def t_ID(t):
    r'[A-Za-z_][\w_]*'
    t.type = reserved_map.get(t.value, 'ID')
    return t


reserved_map = {}
for r in keywords:
    reserved_map[r.lower()] = r

lex.lex(debug=DEBUG_STATUS, optimize=OPTIMIZE)


import ply.yacc as yacc


def p_schema(p):
    '''schema : schema schema_base_elems
              | schema_base_elems
    '''
    if len(p) == 2:
        p[0] = p[1]
    else:
        if isinstance(p[1], list):
            p[0] = p[1] + [p[2]]
        else:
            p[0] = [p[1], p[2]]


def p_schema_base_elems(p):
    '''schema_base_elems : schema_type_element
                         | schema_enum_element
                         | schema_interface_element
                         | schema_def_element
    '''
    p[0] = p[1]


def p_schema_def_element(p):
    '''schema_def_element : SCHEMA schema_type_element_def
    '''
    p[0] = ('SCHEMADEF', 'schema', p[2])


def p_schema_type_element(p):
    '''schema_type_element : TYPE ID schema_type_element_def
                           | TYPE ID IMPLEMENTS ID schema_type_element_def
    '''
    if len(p) == 6:
        p[0] = ('TYPE', p[2], p[4], p[5])
    else:
        p[0] = ('TYPE', p[2], None, p[3])


def p_scheme_type_element_def(p):
    '''schema_type_element_def : LBRACE field_decl_list RBRACE
    '''
    p[0] = p[2]


def p_schema_enum_element(p):
    '''schema_enum_element : ENUM ID LBRACE id_list_no_comma RBRACE
    '''
    p[0] = ('ENUM', p[2], p[4])


def p_schema_interface_element(p):
    '''schema_interface_element : INTERFACE ID schema_type_element_def
    '''
    p[0] = ('INTERFACE', p[2], p[3])


def p_id_list_no_comma(p):
    ''' id_list_no_comma : ID
                         | id_list_no_comma ID
    '''
    if len(p) == 2:
        p[0] = [p[1]]
    else:
        if isinstance(p[1], list):
            p[0] = p[1] + [p[2]]
        else:
            p[0] = [p[1], p[2]]


def p_field_decl_list(p):
    '''field_decl_list : field_decl
                       | field_decl_list field_decl
    '''
    if len(p) == 2:
        p[0] = [p[1]]
    else:
        if isinstance(p[1], list):
            p[0] = p[1] + [p[2]]
        else:
            p[0] = [p[1], p[2]]


def p_field_decl(p):
    '''field_decl : field_name_decl COLON gql_type'''
    p[0] = (p[1], p[3])


def p_field_decl_comma_list(p):
    ''' field_decl_comma_list : field_decl
                              | field_decl_comma_list COMMA field_decl
    '''
    if len(p) == 2:
        p[0] = [p[1]]
    else:
        if isinstance(p[1], list):
            p[0] = p[1] + [p[3]]
        else:
            p[0] = [p[1], p[3]]


def p_field_name_decl(p):
    '''field_name_decl : ID
                       | ID LPAREN field_decl_comma_list RPAREN
    '''
    if len(p) == 2:
        p[0] = ('FIELD', p[1], 'ARGS', [])
    else:
        p[0] = ('FIELD', p[1], 'ARGS', p[3])


def p_gql_type(p):
    '''gql_type : LSQUARE gql_target_type RSQUARE
                | LSQUARE gql_target_type RSQUARE BANG
                | gql_target_type
    '''
    if len(p) == 2:
        p[0] = ('GRAPHQL_ATOM', p[1])
    elif len(p) == 5:
        p[0] = ('GRAPHQL_NON-NULLABLE_LIST', p[2])
    else:
        # its a list type
        p[0] = ('GRAPHQL_LIST', p[2])


def p_gql_target_type(p):
    '''gql_target_type : ID BANG
                       | ID'''
    if len(p) == 3:
        p[0] = ('GRAPHQL_TYPE', 'NON-NULLABLE', p[1])
    else:
        p[0] = ('GRAPHQL_TYPE', 'NULLABLE', p[1])


def p_error(p):
    if not p:
        print('Syntax Error at EOF')
    else:
        print('Syntax Error on {0}'.format(p))


parser = yacc.yacc(
    write_tables=DEBUG_TABLES, debug=PARSER_OUT
)


def parse(data):
    parser.error = 0
    p = parser.parse(data, debug=DEBUG_STATUS)
    if parser.error:
        return None
    return p
